@@ -12,7 +12,7 @@ use Scalar::Util qw(blessed reftype);
 use SHARYANTO::Package::Util qw(package_exists);
 use URI;
 
-our $VERSION = '0.26'; # VERSION
+our $VERSION = '0.27'; # VERSION
 
 our $re_mod = qr/\A[A-Za-z_][A-Za-z_0-9]*(::[A-Za-z_][A-Za-z_0-9]*)*\z/;
 
@@ -554,7 +554,7 @@ sub _pre_tx_action {
         unless $self->{use_tx};
 
     # instantiate custom tx manager, per request if necessary
-    if (reftype($self->{custom_tx_manager}) eq 'CODE') {
+    if ((reftype($self->{custom_tx_manager}) // '') eq 'CODE') {
         eval {
             $self->{_tx} = $self->{custom_tx_manager}->($self);
             die $self->{_tx} unless blessed($self->{_tx});
@@ -745,7 +745,7 @@ Perinci::Access::InProcess - Use Rinci access protocol (Riap) to access Perl cod
 
 =head1 VERSION
 
-version 0.26
+version 0.27
 
 =head1 SYNOPSIS
 
